const pathToRegexp = require('path-to-regexp');
const qs = require('qs');
const { executeMock, registMocks } = require('@atv-fe/mock-api-util');
const merge = require('merge');
const clone = require('clone');
const { MapCache, RequestError, ResponseError } = require('./utils');

const cache = new MapCache({ maxCache: makeRequestFactory.maxCache });

const VALID_FETCH_OPTIONS = ['method', 'mode', 'cache', 'credentials', 'headers', 'redirect', 'referrer', 'body']
/**
 * @typedef FetchOptions
 * @property {object} [params] - path params
 * @property {object} [query] - query string params
 * @property {*}  body
 * 
 * @property {string} [mode] - no-cors, cors, *same-origin
 * @property {string} [cache] -  *default, no-cache, reload, force-cache, only-if-cached
 * @property {string} [credentials] - include, *same-origin, omit
 * @property {string} [redirect] - manual, *follow, error
 * @property {string} [referrer] - no-referrer, *client
 * @property {object} [headers]  
 * 
 * 扩展参数
 * @property {number} [timeout] 
 * @property {number} [responseType]  - *json, string, formData, blob, arrayBuffer
 * @property {boolean} [responseAll=false]  - returns not only body, but also status, headers
 * @property {boolean} [useCache] - where use cache, only works with get method
 * @property {number} [ttl] - lifetime of cache, default 60, in seconds
 * @property {boolean} [debug=false] - debug mode, will print request and response data
 * 
 * API级别扩展参数
 * @property {function} [preFetch] - pre fetch hook
 * @property {function} [postFetch] - pre fetch hook
 * 
 * 全局扩展参数
 * @property {string} [prefix] - url prefix
 * @property {object} [mocks] - api mock
 */

/**
 * 创建请求构造工厂 
 * @param {FetchOptions} globalFetchOptions 
 * @param {function} fetch 
 */

function makeRequestFactory(globalFetchOptions, fetchFn) {
  let { mocks, preFetch: globalPreFetch, postFetch: globalPostFetch, prefix, ...otherGlobalOptions } = globalFetchOptions;
  if (isPlainObject(mocks)) {
    registMocks(mocks);
  }
  const makeRequest = (method, url, apiFetchOptions = {}) => {
    const { preFetch: apiPreFetch, postFetch: apiPostFetch, ...otherApiFetchOptions } = apiFetchOptions;
    const mergeUrlParams = pathToRegexp.compile(url)
    return async (options = {}) => {
      if (!isPlainObject(options)) {
        throw new Error('request options must be literal object');
      }
      options = merge(
        { method: method.toLowerCase() },
        otherGlobalOptions,
        otherApiFetchOptions,
        options
      );

      let urlMergedParams = url;
      if (options.params) {
        urlMergedParams = mergeUrlParams(options.params);
      }
      let fetchUrl = urlMergedParams;
      if (isPlainObject(options.query)) {
        fetchUrl += '?' + qs.stringify(options.query)
      }
      if (fetchUrl[0] === '/' && prefix) {
        fetchUrl = prefix + fetchUrl;
      }

      if (typeof globalPreFetch === 'function') {
        await globalPreFetch(options);
      }
      if (typeof apiPreFetch === 'function') {
        await apiPreFetch(options);
      }

      dealRequestBody(options);

      const fetchOptions = santizeFetchOptions(options);

      if (options.debug) {
        console.log(fetchUrl, JSON.stringify(fetchOptions, null, 2));
      }
      const postFetch = apiPostFetch || globalPostFetch || (v => v);

      const responeAhead = responseData => {
        if (options.debug) {
          console.log(JSON.stringify(responseData, null, 2));
        }
        return postFetch(responseData);
      }
      if (options.useCache && options.method === 'get') {
        let response = cache.get({
          fetchUrl,
          fetchOptions,
          responseAll: options.responseAll || false,
        });
        if (response) {
          return responeAhead(response);
        }
      }
      if (mocks) {
        const [isMock, mockData] = executeMock(urlMergedParams, fetchOptions);
        if (isMock) {
          const responseData = options.responseAll ? mockData : mockData.body;
          return responeAhead(responseData);
        }
      }
      let instance = (fetchFn || fetch)(fetchUrl, fetchOptions);

      if (options.timeout) {
        instance = withTimeout(instance, options.timeout);
      }
      return parseReponse(instance, options).then(responseData => {
        if (options.debug) {
          console.log(JSON.stringify(responseData, null, 2))
        }
        if (options.useCache && options.method === 'get') {
          cache.set(
            {
              fetchUrl,
              fetchOptions,
              responseAll: options.responseAll || false,
            },
            clone(responseData),
            options.ttl || 60000);
        }
        return postFetch(responseData);
      }).catch(err => {
        if (options.debug) {
          console.log(err.message, err.body ? JSON.stringify(err.body) : '');
        }
        const maybePending = postFetch(err);
        const handlerWrapper = wrapper => {
          if (wrapper instanceof Error) {
            throw wrapper;
          }
          return wrapper;
        }
        if (typeof maybePending.then === "function") {
          return maybePending
            .then(handlerWrapper)
            .catch(err => {
              throw err;
            })
        }
        return handlerWrapper(maybePending);
      });
    }
  };
  return makeRequest;
}

function dealRequestBody(options) {
  const { method, headers, body } = options;
  if (!isPlainObject(body) || method === 'get') {
    return;
  }
  const contentType = getContentType(headers);
  if (!contentType || isJSONContentType(contentType)) {
    options.headers = {
      Accept: 'application/json',
      'Content-Type': 'application/json;charset=UTF-8',
      ...options.headers,
    };
    options.body = JSON.stringify(body);
  }
}

function santizeFetchOptions(options) {
  return Object
    .keys(options)
    .filter(key => VALID_FETCH_OPTIONS.indexOf(key) > -1)
    .reduce((a, c) => {
      a[c] = options[c]
      return a;
    }, {})
}

function isPlainObject(obj) {
  return Object.prototype.toString.call(obj) === '[object Object]';
}

function withTimeout(instance, timeout) {
  if (timeout > 0) {
    return Promise.race([
      new Promise((_, reject) =>
        setTimeout(() => reject(new RequestError(`timeout of ${timeout}ms exceeded`)), timeout)
      ),
      instance,
    ]);
  } else {
    return instance;
  }
}

async function parseReponse(instance, options) {
  const { responseAll, responseType = 'json' } = options;
  let response = await instance;
  const body = await response[responseType]();
  if (response.status >= 200 && response.status < 300) {
    if (responseAll) {
      return { status: response.status, headers: response.headers, body };
    }
    return body;
  }
  throw new ResponseError(`http error`, body, response);
}

function getContentType(headers) {
  if (!isPlainObject(headers)) {
    return;
  }
  const key = Object
    .keys(headers)
    .find(key => key.toLowerCase() === 'content-type');

  if (!key) {
    return;
  }
  return headers[key]
}

function isJSONContentType(contentType) {
  return typeof (contentType) === 'string' && /application\/json/.test(contentType);
}

makeRequestFactory.RequestError = RequestError;
makeRequestFactory.ResponseError = ResponseError;

makeRequestFactory.maxCache = 0;
module.exports = makeRequestFactory;